// ============================================================================
// Project:	Deductive Database
// Filename:	pred.cpp
// Purpose:	Predicate Data for Loading Data Files (only EDB predicates)
// Last Change:	16.11.2015
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "strtab.h"
#include "argtype.h"

#include "pred.h"


//=============================================================================
// Constructor and Destructor:
//=============================================================================


//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

pred_c::pred_c(str_t pred_name, int pred_arity)
{
	// Check arguments:
	CHECK(pred_name != STR_NULL, "pred_c::constructor: pred_name is null");
	CHECK(pred_arity >= 0, "pred_c::constructor: pred_arity is negative");
	if(pred_arity > VER_MAX_PRED_ARGS)
		err_c::pred_arity_too_large(pred_name, pred_arity,
						VER_MAX_PRED_ARGS);

	// Initialize attributes:
	Name = pred_name;
	Arity = pred_arity;
	NumRels = 0;
	HashNext = PRED_NULL;

	// Set all argument types and domains to "unknown":
	for(int i = 0; i < VER_MAX_PRED_ARGS; i++) {
		ArgType[i] = ARGTYPE_NULL;
		Dom[i] = STRTAB_NULL;
	}

	// Not really necessary, but maybe safer:
	for(int i = 0; i < PRED_MAX_RELS; i++) {
		Rel[i]  = REL_NULL;
		Bind[i] = BIND_NULL;
	}

	// Set Magic number:
	CHECK_CODE(Magic = PRED_MAGIC);
}


//-----------------------------------------------------------------------------
// Constructor, Special Case for Predicates with one atom/string arg:
//-----------------------------------------------------------------------------

pred_c::pred_c(str_t pred_name, strtab_t dom_1)
{
	// Check arguments:
	CHECK(pred_name != STR_NULL,
			"pred_c::constructor1: pred_name is null");
	CHECK(dom_1 != STRTAB_NULL, "pred_c::constructor1: dom_1 is null");

	// Initialize attributes:
	Name = pred_name;
	Arity = 1;
	NumRels = 0;
	HashNext = PRED_NULL;

	// Check that VER_MAX_PRED_ARGS is large enough:
	CHECK(Arity <= VER_MAX_PRED_ARGS,
			"pred_c::constructor1: VER_MAX_PRED_ARGS too small");

	// Set all argument types and domains to "unknown":
	for(int i = 0; i < VER_MAX_PRED_ARGS; i++) {
		ArgType[i] = ARGTYPE_NULL;
		Dom[i] = STRTAB_NULL;
	}

	// Set data for predicate argument:
	ArgType[0] = ARGTYPE_ATOM_STR;
	Dom[0] = dom_1;

	// Not really necessary, but maybe safer:
	for(int i = 0; i < PRED_MAX_RELS; i++) {
		Rel[i]  = REL_NULL;
		Bind[i] = BIND_NULL;
	}

	// Set Magic number:
	CHECK_CODE(Magic = PRED_MAGIC);
}

//-----------------------------------------------------------------------------
// Constructor, Special Case for Predicates with two atom/string args:
//-----------------------------------------------------------------------------

pred_c::pred_c(str_t pred_name, strtab_t dom_1, strtab_t dom_2)
{
	// Check arguments:
	CHECK(pred_name != STR_NULL,
			"pred_c::constructor2: pred_name is null");
	CHECK(dom_1 != STRTAB_NULL, "pred_c::constructor2: dom_1 is null");
	CHECK(dom_1 != STRTAB_NULL, "pred_c::constructor2: dom_2 is null");

	// Initialize attributes:
	Name = pred_name;
	Arity = 2;
	NumRels = 0;
	HashNext = PRED_NULL;

	// Check that VER_MAX_PRED_ARGS is large enough:
	CHECK(Arity <= VER_MAX_PRED_ARGS,
			"pred_c::constructor2: VER_MAX_PRED_ARGS too small");

	// Set all argument types and domains to "unknown":
	for(int i = 0; i < VER_MAX_PRED_ARGS; i++) {
		ArgType[i] = ARGTYPE_NULL;
		Dom[i] = STRTAB_NULL;
	}

	// Set data for predicate arguments:
	ArgType[0] = ARGTYPE_ATOM_STR;
	ArgType[1] = ARGTYPE_ATOM_STR;
	Dom[0] = dom_1;
	Dom[1] = dom_2;

	// Not really necessary, but maybe safer:
	for(int i = 0; i < PRED_MAX_RELS; i++) {
		Rel[i]  = REL_NULL;
		Bind[i] = BIND_NULL;
	}

	// Set Magic number:
	CHECK_CODE(Magic = PRED_MAGIC);
}


//-----------------------------------------------------------------------------
// Constructor, Special Case for Predicates with one numeric arg:
//-----------------------------------------------------------------------------

pred_c::pred_c(str_t pred_name, argtype_t argtype_1)
{
	// Check arguments:
	CHECK(pred_name != STR_NULL,
			"pred_c::constructor3: pred_name is null");
	CHECK(argtype_valid(argtype_1),
			"pred_c::constructor3: argtype is invalid");
	CHECK(argtype_1 != ARGTYPE_NULL,
			"pred_c::constructor3: argtype is null");
	CHECK(argtype_1 != ARGTYPE_ATOM && argtype_1 != ARGTYPE_STR &&
			argtype_1 != ARGTYPE_ATOM_STR,
			"pred_c::constructor3: argtype needs domain");

	// Initialize attributes:
	Name = pred_name;
	Arity = 1;
	NumRels = 0;
	HashNext = PRED_NULL;

	// Check that VER_MAX_PRED_ARGS is large enough:
	CHECK(Arity <= VER_MAX_PRED_ARGS,
			"pred_c::constructor3: VER_MAX_PRED_ARGS too small");

	// Set all argument types and domains to "unknown":
	for(int i = 0; i < VER_MAX_PRED_ARGS; i++) {
		ArgType[i] = ARGTYPE_NULL;
		Dom[i] = STRTAB_NULL;
	}

	// Set data for predicate argument:
	ArgType[0] = argtype_1;

	// Not really necessary, but maybe safer:
	for(int i = 0; i < PRED_MAX_RELS; i++) {
		Rel[i]  = REL_NULL;
		Bind[i] = BIND_NULL;
	}

	// Set Magic number:
	CHECK_CODE(Magic = PRED_MAGIC);
}

//-----------------------------------------------------------------------------
// Constructor, Special Case for Predicates with two numeric args:
//-----------------------------------------------------------------------------

pred_c::pred_c(str_t pred_name, argtype_t argtype_1, argtype_t argtype_2)
{
	// Check arguments:
	CHECK(pred_name != STR_NULL,
			"pred_c::constructor4: pred_name is null");
	CHECK(argtype_valid(argtype_1),
			"pred_c::constructor4: argtype_1 is invalid");
	CHECK(argtype_1 != ARGTYPE_NULL,
			"pred_c::constructor4: argtype_1 is null");
	CHECK(argtype_1 != ARGTYPE_ATOM && argtype_1 != ARGTYPE_STR &&
			argtype_1 != ARGTYPE_ATOM_STR,
			"pred_c::constructor4: argtype_1 needs domain");
	CHECK(argtype_valid(argtype_2),
			"pred_c::constructor4: argtype_2 is invalid");
	CHECK(argtype_2 != ARGTYPE_NULL,
			"pred_c::constructor4: argtype_2 is null");
	CHECK(argtype_2 != ARGTYPE_ATOM && argtype_2 != ARGTYPE_STR &&
			argtype_2 != ARGTYPE_ATOM_STR,
			"pred_c::constructor4: argtype_2 needs domain");

	// Initialize attributes:
	Name = pred_name;
	Arity = 2;
	NumRels = 0;
	HashNext = PRED_NULL;

	// Check that VER_MAX_PRED_ARGS is large enough:
	CHECK(Arity <= VER_MAX_PRED_ARGS,
			"pred_c::constructor4: VER_MAX_PRED_ARGS too small");

	// Set all argument types and domains to "unknown":
	for(int i = 0; i < VER_MAX_PRED_ARGS; i++) {
		ArgType[i] = ARGTYPE_NULL;
		Dom[i] = STRTAB_NULL;
	}

	// Set data for predicate arguments:
	ArgType[0] = argtype_1;
	ArgType[1] = argtype_2;

	// Not really necessary, but maybe safer:
	for(int i = 0; i < PRED_MAX_RELS; i++) {
		Rel[i]  = REL_NULL;
		Bind[i] = BIND_NULL;
	}

	// Set Magic number:
	CHECK_CODE(Magic = PRED_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

pred_c::~pred_c()
{
	// Check this object:
	CHECK_VALID("pred_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Object Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// add_rel: Add relation to store facts of this predicate.
//-----------------------------------------------------------------------------

bool pred_c::add_rel(rel_t rel, bind_t bind)
{
	// Check this object:
	CHECK_VALID("pred_c::add_rel");

	// Check parameter:
	CHECK_PAR(rel, "pred_c::add_rel");
	CHECK(bind_valid(bind), "pred_c::add_rel: Invalid binding pattern");
	CHECK(bind_arity(bind) == Arity,
			"pred_c::add_rel: Binding pattern != Arity");

	// Is there still space to store a relation?
	if(NumRels == PRED_MAX_RELS) {
		err_c::pred_too_many_rels(Name, PRED_MAX_RELS);
		return false;
	}

	// Store relation and binding pattern:
	Rel[NumRels] = rel;
	Bind[NumRels] = bind;
	NumRels++;
	return true;
}


//-----------------------------------------------------------------------------
// store_fact: Store fact in all relations attached to this predicate.
//-----------------------------------------------------------------------------

bool pred_c::store_fact(int args[])
{
	for(int i = 0; i < NumRels; i++) {
		bool ok;
		switch(Bind[i]) {
		case BIND_B:
			CHECK(Arity == 1,
				"pred_c::store_fact: Binding b != arity");
			ok = static_cast<set_1_t>(Rel[i]) ->
				insert(args[0]);
			break;
		case BIND_F:
			CHECK(Arity == 1,
				"pred_c::store_fact: Binding f != arity");
			ok = static_cast<list_1_t>(Rel[i]) ->
				insert(args[0]);
			break;
		case BIND_BB:
			CHECK(Arity == 2,
				"pred_c::store_fact: Binding bb != arity");
			ok = static_cast<set_2_t>(Rel[i]) ->
				insert(args[0], args[1]);
			break;
		case BIND_FF:
			CHECK(Arity == 2,
				"pred_c::store_fact: Binding ff != arity");
			ok = static_cast<list_2_t>(Rel[i]) ->
				insert(args[0], args[1]);
			break;
		case BIND_BF:
			CHECK(Arity == 2,
				"pred_c::store_fact: Binding bf != arity");
			ok = static_cast<rel_n_n_t>(Rel[i]) ->
				insert(args[0], args[1]);
			break;
		case BIND_FB:
			CHECK(Arity == 2,
				"pred_c::store_fact: Binding fb != arity");
			ok = static_cast<rel_n_n_t>(Rel[i]) ->
				insert(args[1], args[0]);
			break;
		default:
			CHECK_IMPOSSIBLE(
				"pred_c::store_fact: Binding pattern missing");
			ok = false;
		}
		if(!ok)
			return false;
	}

	// No error detected:
	return true;
}

//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Integrity Check.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	pred_c::check() const
{
	// Check magic number:
	if(Magic != PRED_MAGIC)
		return "wrong magic number";

	// Predicate name must be not null:
	if(Name == STR_NULL)
		return "predicate name is null";

	// Arity must be non-negative:
	if(Arity < 0)
		return "predicate arity is negative";
	// Note: Arity can be > VER_MAX_PRED_ARGS.
	// An error message is generated via err_c in this case.
	// All calls with an argument number >= VER_MAX_PRED_ARGS
	// must be ignored.

	// Argument types must be valid:
	for(int i = 0; i < Arity && i < VER_MAX_PRED_ARGS; i++)
		if(!argtype_valid(ArgType[i]))
			return "predicate has invalid argument type";

	// Check interval for NumRels:
	if(NumRels < 0)
		return "NumRels is negative";
	if(NumRels > PRED_MAX_RELS)
		return "NumRels is too large";

	// NumRels Relations must be defined:
	for(int i = 0; i < NumRels; i++)
		if(Rel[i] == REL_NULL)
			return "Undefined relation in EDB predicate";

	// Binding pattern must be defined for each relation:
	for(int i = 0; i < NumRels; i++)
		if(Bind[i] == BIND_NULL)
			return "Undefined binding pattern in EDB predicate";

	// Ok:
	return 0;
}

#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void	pred_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "Predicate Data for Fact Loader";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("Name:       '");
	err_c::dump_str(Name);
	err_c::dump_str("'");
	err_c::dump_nl();

	err_c::dump_str("Arity:       ");
	err_c::dump_int(Arity);
	err_c::dump_nl();

	err_c::dump_str("NumRels:     ");
	err_c::dump_int(NumRels);
	err_c::dump_nl();

	err_c::dump_nl();

	// Show argument types:
	err_c::dump_str("Argument Types:");
	err_c::dump_nl();
	for(int i = 0; i < Arity && i < VER_MAX_PRED_ARGS; i++) {
		err_c::dump_str("    ");
		err_c::dump_int(i);
		err_c::dump_str(". ");
		err_c::dump_str(argtype_name(ArgType[i]));
		err_c::dump_nl();
	}
	err_c::dump_nl();

	// Show domains for arguments:
	err_c::dump_str("Domains for Predicate Arguments:");
	err_c::dump_nl();
	for(int i = 0; i < Arity; i++) {
		err_c::dump_str("  Arg. ");
		err_c::dump_str(str_fill(i, 2));
		err_c::dump_int(i);
		err_c::dump_str(": ");
		if(Dom[i] == STRTAB_NULL)
			err_c::dump_str("*** NULL ***");
		else
			err_c::dump_str(Dom[i]->name());
		err_c::dump_nl();
	}
	err_c::dump_nl();

	// Show relations for storing predicate facts:
	if(NumRels > 0) {
		err_c::dump_str("Relations for Storing Predicate Facts:");
		err_c::dump_nl();
		for(int i = 0; i < NumRels; i++) {
			err_c::dump_str("  Rel. ");
			err_c::dump_str(str_fill(i, 2));
			err_c::dump_int(i);
			err_c::dump_str(":");
			if(Rel[i] == REL_NULL)
				err_c::dump_str("*** NULL ***");
			else
				err_c::dump_str(Rel[i]->name());
			err_c::dump_str(" (Binding pattern: '");
			err_c::dump_str(bind_name(Bind[i]));
			err_c::dump_str("')");
			err_c::dump_nl();
		}
		err_c::dump_nl();
	}

	// Done:
	err_c::dump_end();
}
#endif

